@@ -12,15 +12,11 @@
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
 #include "pixel.h"
 
-pixel_t pixel_get(int x, int y, const unsigned char* pix, unsigned slice, unsigned pixel, unsigned dx, unsigned dy, int opt_tes)
+pixel_t pixel_get(int x, int y, const unsigned char* pix, unsigned slice, unsigned pixel, int dx, int dy, int opt_tes)
 {
 	const unsigned char* p;
 	unsigned i;
@@ -49,20 +45,26 @@ pixel_t pixel_get(int x, int y, const unsigned char* pix, unsigned slice, unsign
 	p = pix + (y * slice) + (x * pixel);
 
 	v = 0;
-	for(i=0;i<pixel;++i)
-		v |= ((pixel_t)p[i]) << (i*8);
+	for (i = 0; i < pixel; ++i)
+		v |= ((pixel_t)p[i]) << (i * 8);
 
 	return v;
 }
 
-void pixel_put(int x, int y, unsigned char* pix, unsigned slice, unsigned pixel, unsigned dx, unsigned dy, pixel_t v)
+void pixel_put(int x, int y, unsigned char* pix, unsigned slice, unsigned pixel, int dx, int dy, pixel_t v)
 {
 	unsigned char* p;
 	unsigned i;
 
+	if (x < 0 || x >= dx)
+		return;
+	if (y < 0 || y >= dy)
+		return;
+
 	p = pix + (y * slice) + (x * pixel);
 
-	for(i=0;i<pixel;++i) {
-		p[i] = v >> (i*8);
+	for (i = 0; i < pixel; ++i) {
+		p[i] = v >> (i * 8);
 	}
 }
+
